import 'dart:typed_data';

import 'package:collection/collection.dart';
import 'package:meta/meta.dart';

import 'raw_reader.dart';
import 'raw_writer.dart';

/// Something that implements both [SelfEncoder] and [SelfDecoder].
abstract class SelfCodec extends SelfEncoder with SelfDecoder {}

/// Something that can decode itself using [RawReader].
abstract class SelfDecoder {
  /// Decodes state from the bytes.
  void decodeSelf(RawReader reader);

  /// Decodes state from the SelfEncoder.
  void decodeSelfFromSelfEncoder(SelfEncoder value) {
    final byteData = value.toImmutableByteData();
    final reader = new RawReader.withByteData(byteData);
    decodeSelf(reader);
  }
}

/// Something that can encode itself using [RawWriter].
abstract class SelfEncoder {
  const SelfEncoder();

  /// Determines hash by serializing this value.
  @override
  int get hashCode => const SelfEncoderEquality().hash(this);

  /// Determines equality by serializing both values.
  @override
  bool operator ==(other) {
    return other is SelfEncoder &&
        const SelfEncoderEquality().equals(this, other);
  }

  /// Encodes this object.
  void encodeSelf(RawWriter writer);

  /// Returns recommended capacity for a buffer.
  @visibleForTesting
  int encodeSelfCapacity() => 64;

  /// Returns an immutable encoding of this value.
  ByteData toImmutableByteData() => toMutableByteData();

  /// Returns an immutable encoding of this value.
  List<int> toImmutableUint8List() => toMutableUint8List();

  /// Returns a mutable encoding of this value.
  ByteData toMutableByteData() {
    final capacity = encodeSelfCapacity();
    final writer = new RawWriter.withCapacity(capacity);
    encodeSelf(writer);
    return writer.toByteDataView();
  }

  /// Returns a mutable encoding of this value.
  List<int> toMutableUint8List() {
    final capacity = encodeSelfCapacity();
    final writer = new RawWriter.withCapacity(capacity);
    encodeSelf(writer);
    return writer.toUint8ListView();
  }
}

/// Equality for [SelfEncoder].
///
/// Used by '==' and 'hashCode' in [SelfEncoder].
class SelfEncoderEquality implements Equality<SelfEncoder> {
  const SelfEncoderEquality();

  @override
  bool equals(SelfEncoder e1, SelfEncoder e2) {
    final bytes = e1.toImmutableUint8List();
    final otherBytes = e2.toImmutableUint8List();
    if (bytes.length != otherBytes.length) {
      return false;
    }
    for (var i = 0; i < bytes.length; i++) {
      if (bytes[i] != otherBytes[i]) {
        return false;
      }
    }
    return true;
  }

  @override
  int hash(SelfEncoder e) {
    final bytes = e.toImmutableUint8List();
    var result = 0;
    for (var i = 0; i < bytes.length; i++) {
      result ^= bytes[i] << ((i * 7) % 24);
    }
    return result;
  }

  @override
  bool isValidKey(Object o) => o is SelfEncoder;
}
